module jp.ku6.wallpaper.wallpaper;

import std.stream;
import jp.ku6.windows.win32api.windows;
import jp.ku6.windows.directx.dxgiformat;
import jp.ku6.windows.directx.dxgitype;
import jp.ku6.windows.directx.dxgi;
import jp.ku6.windows.directx.d3dcommon;
import jp.ku6.windows.directx.d3d10;
import jp.ku6.windows.directx.d3d10misc;
import jp.ku6.windows.directx.d3d10_1;
import jp.ku6.windows.directx.d3dcompiler;

class WallPaper
{
	/++
	 + コンストラクタ
	 +/
	this(HWND targetWindowHandle)
	{
		//--------------------------------------------------------------
		// DXGI_ADAPTER_DESC1.AdapterLuidを取得

		// DXGIのDLLをロード
		dxgiDll_ = LoadLibrary("dxgi.dll\0".ptr);
		if (dxgiDll_ is null)
			throw new Error("Failed load dxgi.dll");

		// DLLからCreateDXGIFactory1を取得
		CreateDXGIFactory1Function CreateDXGIFactory1 = cast(CreateDXGIFactory1Function)GetProcAddress(dxgiDll_, "CreateDXGIFactory1\0".ptr);
		if (CreateDXGIFactory1 is null)
		{
			cleanup();
			throw new Error("Failed get CreateDXGIFactory1 function");
		}

		// CreateDXGIFactory1を使ってIDXGIFactory1を取得
		if (FAILED(CreateDXGIFactory1(&IID_IDXGIFactory, cast(void**)(&dxgiFactory_))))
		{
			cleanup();
			throw new Error("Failed create IDXGIFactory1");
		}

		// IDXGIFactory1を使ってIDXGIAdapter1を列挙(といっても最初の1個だけ)
		if (dxgiFactory_.EnumAdapters1(0, &dxgiAdapter_) == DXGI_ERROR_NOT_FOUND) 
		{ 
			cleanup();
			throw new Error("Failed enum adapter");
		}

		// IDXGIAdapter1からDXGI_ADAPTER_DESC1を取得
		DXGI_ADAPTER_DESC1 adapterDesc;
		if (FAILED(dxgiAdapter_.GetDesc1(&adapterDesc)))
		{
			cleanup();
			throw new Error("Failed get adapterd desc");
		}


		//--------------------------------------------------------------
		// DXGI_ADAPTER_DESC1.AdapterLuidを使って共有リソースのハンドルを取得

		// DWMAPIのDLLをロード
		dwmapiDll_ = LoadLibrary("dwmapi.dll\0".ptr);
		if (dwmapiDll_ is null)
		{
			cleanup();
			throw new Error("Failed load dwmapi.dll");
		}

		// 隠しAPIを取得(index=100)
		DwmGetSharedSurfaceFunction DwmGetSharedSurface = cast(DwmGetSharedSurfaceFunction)GetProcAddress(dwmapiDll_, cast(char*)100);
		if (DwmGetSharedSurface is null)
		{
			cleanup();
			throw new Error("Failed get DwmGetSharedSurface function");
		}

		// 隠しAPIを取得を使って共有リソースのハンドルを取得
		DXGI_FORMAT format;
		HANDLE sharedHandle;
		jp.ku6.windows.directx.dxgi.LUID nullLUID;
		if (FAILED(DwmGetSharedSurface(targetWindowHandle, adapterDesc.AdapterLuid, nullLUID, &format, &sharedHandle, null)))
		{
			cleanup();
			throw new Error("Failed get DwmGetSharedSurface function");
		}

		// 取れているかチェック
		if (sharedHandle is null)
		{
			cleanup();
			throw new Error("Failed get sharedHandle");
		}


		//--------------------------------------------------------------
		// 共有リソースのハンドルを使ってID3D10Texture2Dを取得

		// Direct 3D 10.1のDLLをロード
		d3d10_1Dll_ = LoadLibrary("d3d10_1.dll\0".ptr);
		if (d3d10_1Dll_ is null)
		{
			cleanup();
			throw new Error("Failed load d3d10_1.dll");
		}

		// D3D10CreateDevice1を取得
		D3D10CreateDevice1Function D3D10CreateDevice1 = cast(D3D10CreateDevice1Function)GetProcAddress(d3d10_1Dll_, "D3D10CreateDevice1\0".ptr);
		if (D3D10CreateDevice1 is null)
		{
			cleanup();
			throw new Error("Failed get D3D10CreateDevice1 function");
		}

		// D3D10CreateDevice1を使用してID3D10Device1を取得
		if (FAILED(D3D10CreateDevice1(dxgiAdapter_, D3D10_DRIVER_TYPE.D3D10_DRIVER_TYPE_HARDWARE, null, D3D10_CREATE_DEVICE_FLAG.D3D10_CREATE_DEVICE_DEBUG, D3D10_FEATURE_LEVEL1.D3D10_FEATURE_LEVEL_10_1, D3D10_1_SDK_VERSION, &device_)))
		{
			cleanup();
			throw new Error("Failed create device");
		}

		// 共有リソースのハンドルからID3D10Resourceを取得
		ID3D10Resource resource;
		HRESULT hr = device_.OpenSharedResource(sharedHandle, &IID_ID3D10Resource, cast(void**)&resource);
		if (FAILED(hr))
		{
			cleanup();
			throw new Error("Failed open shared resource: " ~ std.conv.to!string(hr));
		}

		// ID3D10ResourceをID3D10Texture2Dにする
		if (FAILED(resource.QueryInterface(&IID_ID3D10Texture2D, cast(void**)&renderTarget_)))
		{
			resource.Release();
			cleanup();
			throw new Error("Failed query interface");
		}
		resource.Release();

/*
		// 対象のテクスチャを1色で塗りつぶし

		if (FAILED(device_.CreateRenderTargetView(renderTarget_, null, &renderTargetView_)))
		{
			cleanup();
			throw new Error("Failed create depth stencil shadow texture view");
		}

		float[4] clearColor = [1, 0, 1, 1];
		device_.ClearRenderTargetView(renderTargetView_, clearColor.ptr);
		device_.Flush();
*/

		//--------------------------------------------------------------
		// ID3D10Texture2DをBMPファイルとして保存する

		// 対象のテクスチャ情報を取得
		D3D10_TEXTURE2D_DESC textureDesc;
		renderTarget_.GetDesc(&textureDesc);

		// 同じフォーマットでCPUアクセス可能なキャプチャ用テクスチャを作成
		ID3D10Texture2D captureTexture;
		D3D10_TEXTURE2D_DESC captureTextureDesc = textureDesc;
		captureTextureDesc.Usage = D3D10_USAGE.D3D10_USAGE_STAGING;
		captureTextureDesc.BindFlags = 0;
		captureTextureDesc.CPUAccessFlags = D3D10_CPU_ACCESS_FLAG.D3D10_CPU_ACCESS_READ;
		if (FAILED(device_.CreateTexture2D(&captureTextureDesc, null, &captureTexture)))
		{
			cleanup();
			throw new Error("Failed create capture texture");
		}

		// 対象のテクスチャからキャプチャ用テクスチャにデータをコピー
		device_.CopyResource(captureTexture, renderTarget_);


		// キャプチャ用テクスチャをロック
		D3D10_MAPPED_TEXTURE2D mapped;
		if (FAILED(captureTexture.Map(D3D10CalcSubresource(0, 0, 1), D3D10_MAP.D3D10_MAP_READ, 0, &mapped)))
		{
			captureTexture.Release();
			cleanup();
			throw new Error("Failed map texture");
		}

		// BMPファイルとして保存
		auto file = new BufferedFile("capture.bmp", FileMode.OutNew);
		file.write(0x2c364d42);
		file.write(0x00000022);
		file.write(0x00360000);
		file.write(0x00280000);
		file.write(cast(ushort)0);
		file.write(textureDesc.Width);
		file.write(textureDesc.Height);
		file.write(0x00200001);
		file.write(0x00000000);
		file.write(textureDesc.Width * 4 * textureDesc.Height);
		file.write(0x00000000);
		file.write(0x00000000);
		file.write(0x00000000);
		file.write(0x00000000);

		uint* bitmap = cast(uint*)mapped.pData;
		for (uint y; y < textureDesc.Height; ++y)
		{
			for (uint x; x < textureDesc.Width; ++x)
			{
				file.write(bitmap[textureDesc.Width * (textureDesc.Height - y - 1) + x]);
			}
		}

		file.close();

		// キャプチャ用テクスチャをアンロック
		captureTexture.Unmap(D3D10CalcSubresource(0, 0, 1));

		captureTexture.Release();
	}

	/++
	 + デストラクタ
	 +/
	~this()
	{
		cleanup();
	}

private:

	extern (Windows)
	alias HRESULT function(HWND handle, jp.ku6.windows.directx.dxgi.LUID adapterLuid, jp.ku6.windows.directx.dxgi.LUID someLuid, DXGI_FORMAT* format, HANDLE* sharedHandle, ulong* unknown) DwmGetSharedSurfaceFunction;

	HMODULE dxgiDll_;

	HMODULE dwmapiDll_;

	HMODULE d3d10_1Dll_;

	IDXGIFactory1 dxgiFactory_;

	IDXGIAdapter1 dxgiAdapter_;

	ID3D10Device1 device_;

	ID3D10Texture2D renderTarget_;

	ID3D10RenderTargetView renderTargetView_;

	/+
	 + 解放処理
	 +/
	void cleanup()
	{
		if (renderTargetView_ !is null)
		{
			renderTargetView_.Release();
			renderTargetView_ = null;
		}

		if (renderTarget_ !is null)
		{
			renderTarget_.Release();
			renderTarget_ = null;
		}

		if (device_ !is null)
		{
			device_.Release();
			device_ = null;
		}

		if (dxgiAdapter_ !is null)
		{
			dxgiAdapter_.Release();
			dxgiAdapter_ = null;
		}

		if (dxgiFactory_ !is null)
		{
			dxgiFactory_.Release();
			dxgiFactory_ = null;
		}

		if (d3d10_1Dll_ !is null)
		{
			FreeLibrary(d3d10_1Dll_);
			d3d10_1Dll_ = null;
		}

		if (dwmapiDll_ !is null)
		{
			FreeLibrary(dwmapiDll_);
			dwmapiDll_ = null;
		}

		if (dxgiDll_ !is null)
		{
			FreeLibrary(dxgiDll_);
			dxgiDll_ = null;
		}
	}
}

